from dataclasses import dataclass


@dataclass
class Quantity:
    """Base quantity class"""

    cpu: str
    memory: str


@dataclass
class Limits(Quantity):
    """Limits quantity"""

    cpu: str = "0.15"
    memory: str = "300Mi"


@dataclass
class Requests(Quantity):
    """Requests quantity"""

    cpu: str = "0.1"
    memory: str = "200Mi"



from typing import Union

from pydantic import Field, TypeAdapter
from typing_extensions import Annotated

from .cron_workflow import CronWorkflow
from .workflow import Workflow

# https://docs.pydantic.dev/latest/usage/types/unions/#nested-discriminated-unions
BaseWorkflow = Annotated[
    Union[
        CronWorkflow,
        Workflow,
    ],
    Field(discriminator="kind"),
]


# See: https://github.com/pydantic/pydantic/discussions/4950#discussioncomment-5775767
def WorkflowFactory(data):
    return TypeAdapter(BaseWorkflow).validate_python(data)



"""
Source: https://argoproj.github.io/argo-workflows/fields/#dagtask
"""

from typing import List, Optional

from pydantic import BaseModel

from .arguments import Arguments


class DAGTask(BaseModel):
    """DAGTask represents a node in the graph during DAG execution"""

    name: str
    template: str
    arguments: Optional[Arguments] = None
    dependencies: List[str] = []



from typing import Literal

from pydantic import BaseModel

from .cron_workflow_spec import CronWorkflowSpec
from .metadata import Metadata
from .workflow_spec import WorkflowSpec


class CronWorkflow(BaseModel):
    """CronWorkflow model"""

    kind: Literal["CronWorkflow"]
    metadata: Metadata
    spec: CronWorkflowSpec

    def set_workflow_spec(self, workflow_spec: WorkflowSpec):
        self.spec.workflowSpec = workflow_spec
        return self



# pylint: disable=line-too-long,too-many-arguments,too-many-locals
"""CLI methods to run dbt-argo app"""

import sys

import click
import yaml
from dbt.cli.main import dbtRunner, dbtRunnerResult
from dbt.contracts.graph.manifest import Manifest

from .submitter import ArgoSubmitter  # pylint: disable=import-error
from .utils import copy_dir_to_gcs
from .workflows import create_dbt_cronworkflow, create_dbt_workflow

# TODO: find a way not to repeat click.options


@click.group(invoke_without_command=True)
def cli():
    """Test"""
    ctx = click.get_current_context()
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@cli.command(name="create")
# Same for cron
@click.option("--name", default="analytics-adhoc", show_default=True)
@click.option("--gcs-bucket", envvar="GCS_BUCKET", required=True, show_default=True)
@click.option("--namespace", envvar="K8S_NAMESPACE", show_default=True)
@click.option("--service-account-name", envvar="K8S_SA", show_default=True)
@click.option("--env", envvar="ENV", show_default=True)
@click.option("--gcp_project_id", envvar="GCP_PROJECT_ID", show_default=True)
@click.option("--submit", is_flag=True)
@click.option("--echo", is_flag=True)
@click.argument("command", type=click.Choice(["build", "run"]))
def create_workflow(
    name,
    gcs_bucket,
    namespace,
    service_account_name,
    env,
    gcp_project_id,
    submit,
    echo,
    command,
):
    """Create a dbt Argo Workflow"""

    # https://docs.getdbt.com/reference/programmatic-invocations#reusing-objects
    res: dbtRunnerResult = dbtRunner().invoke(["parse"])
    if isinstance(res.result, Manifest):
        manifest = res.result

        workflow = create_dbt_workflow(
            manifest,
            name=name,
            command=command,
            service_account_name=service_account_name,
            namespace=namespace,
            gcs_bucket=gcs_bucket,
            env=env,
            gcp_project_id=gcp_project_id,
        )

        filepath = workflow.to_file(name=name)

        if echo:
            click.echo()
            click.echo(workflow.to_yaml())

        if submit:
            submit_workflow(filepath, dry_run=False)


@cli.group(invoke_without_command=True)
def cron():
    """Manage cron workflow"""
    ctx = click.get_current_context()
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@cron.command(name="create")
# Workfloc command
@click.option("--name", default="analytics-adhoc", show_default=True)
@click.option("--gcs-bucket", envvar="GCS_BUCKET", required=True, show_default=True)
@click.option("--namespace", envvar="K8S_NAMESPACE", show_default=True)
@click.option("--service-account-name", envvar="K8S_SA", show_default=True)
@click.option("--env", envvar="ENV", show_default=True)
@click.option("--gcp_project_id", envvar="GCP_PROJECT_ID", show_default=True)
# Cron specific
@click.option("--schedule", default="0 4 * * *", show_default=True)
@click.option("--timezone", default="Europe/Paris", show_default=True)
@click.option("--suspend", default=False, show_default=True)
# submit
@click.option("--submit", is_flag=True)
@click.option("--echo", is_flag=True)
@click.argument("command", type=click.Choice(["build", "run"]))
def create_cron(
    name: str,
    gcs_bucket: str,
    namespace: str,
    service_account_name: str,
    env: str,
    gcp_project_id: str,
    schedule: str,
    timezone: str,
    suspend: bool,
    submit,
    echo,
    command: str,
):
    """Create a dbt Argo CronWorkflow"""

    # https://docs.getdbt.com/reference/programmatic-invocations#reusing-objects
    res: dbtRunnerResult = dbtRunner().invoke(["parse"])
    if isinstance(res.result, Manifest):
        manifest = res.result

        workflow = create_dbt_cronworkflow(
            manifest,
            schedule=schedule,
            name=name,
            command=command,
            service_account_name=service_account_name,
            namespace=namespace,
            gcs_bucket=gcs_bucket,
            env=env,
            gcp_project_id=gcp_project_id,
            timezone=timezone,
            suspend=suspend,
        )

        filepath = workflow.to_file(name=name)

        if echo:
            click.echo()
            click.echo(workflow.to_yaml())

        if submit:
            submit_workflow(filepath, dry_run=False)


@cli.command(name="submit")
@click.option("-f", "--filename")
@click.option("-n", "--namespace", default="default")
@click.option("--dry-run", is_flag=True)
def submit_workflow(filename, namespace, dry_run):
    """dbt-argo deploy {FILE}"""
    click.echo(f" Deploying {filename}!")

    submitter = ArgoSubmitter(namespace=namespace)

    with open(filename, encoding="utf8") as file:
        workflow_yaml = yaml.safe_load(file)

    if dry_run:
        sys.exit("Dry run mode...")

    submitter.submit(workflow_yaml)


@cli.command(name="cp")
@click.argument("folder", default=".", required=True, type=str)
@click.argument("gcs_path", required=True, type=str)
def cp(folder, gcs_path):
    """copy local folder to GCS Bucket"""
    copy_dir_to_gcs(folder, gcs_path)



# Source: https://github.com/couler-proj/couler/blob/master/couler/argo_submitter.py
import logging
import re

import pyaml
import yaml
from kubernetes import client as k8s_client
from kubernetes import config


class WorkflowCRD(object):
    PLURAL = "workflows"
    KIND = "Workflow"
    GROUP = "argoproj.io"
    VERSION = "v1alpha1"
    NAME_MAX_LENGTH = 45
    NAME_PATTERN = r"[a-z]([-a-z0-9]*[a-z0-9])?"


class CronWorkflowCRD(WorkflowCRD):
    PLURAL = "cronworkflows"
    KIND = "CronWorkflow"


class _SubmitterImplTypes:
    PYTHON = "Python"
    GO = "Go"


# TODO: some k8s common parts can move to another file later.
class ArgoSubmitter:
    """A submitter which submits a workflow to Argo"""

    _default_submitter = None

    def __init__(
        self,
        namespace="default",
        config_file=None,
        context=None,
        client_configuration=None,
        persist_config=True,
    ):
        logging.basicConfig(level=logging.INFO)
        self.namespace = namespace
        logging.info("Argo submitter namespace: %s" % self.namespace)

        try:
            config.load_kube_config(
                config_file, context, client_configuration, persist_config
            )
            logging.info(
                "Found local kubernetes config. " "Initialized with kube_config."
            )
            if client_configuration is not None:
                logging.info("Setting default k8s client config as provided")
                k8s_client.Configuration.set_default(client_configuration)
        except Exception:
            logging.info("Cannot find local k8s config. " "Trying in-cluster config.")
            config.load_incluster_config()
            logging.info("Initialized with in-cluster config.")

        self._custom_object_api_client = k8s_client.CustomObjectsApi()
        self._core_api_client = k8s_client.CoreV1Api()

    @staticmethod
    def check_name(name):
        """Check the name is valid or not"""
        if len(name) > WorkflowCRD.NAME_MAX_LENGTH:
            raise ValueError(
                "Name is too long. Max length: {}, now: {}"
                "".format(WorkflowCRD.NAME_MAX_LENGTH, len(name))
            )
        if "." in name:
            raise ValueError("Name cannot include dot.")
        if "_" in name:
            raise ValueError("Name cannot include underscore.")

        match_obj = re.match(WorkflowCRD.NAME_PATTERN, name)
        if not match_obj:
            raise ValueError(
                "Name is invalid. Regex used for validation is %s"
                % WorkflowCRD.NAME_PATTERN
            )

    def get_custom_object_api_client(self):
        return self._custom_object_api_client

    def get_core_api_client(self):
        return self._core_api_client

    def submit(self, workflow_yaml, secrets=None):
        wf_name = (
            workflow_yaml["metadata"]["name"]
            if "name" in workflow_yaml["metadata"]
            else workflow_yaml["metadata"]["generateName"]
        )
        if secrets:
            for secret in secrets:
                if secret.use_existing is False:
                    if secret.artifact_secret is True:
                        # For artifact secrets, check if the secret already exists
                        if secret.name in [
                            x.metadata.name
                            for x in self.get_core_api_client()
                            .list_namespaced_secret(namespace=self.namespace)
                            .items
                        ]:
                            logging.info(
                                "Secret {} already exists in the {} namespace. Skipping creation.".format(
                                    secret.name, self.namespace
                                )
                            )
                        else:
                            # Otherwise create the secret
                            self._create_secret(secret.to_yaml())
                    else:
                        # For all other secrets
                        self._create_secret(secret.to_yaml())

        logging.info("Checking workflow name/generatedName %s" % wf_name)
        self.check_name(wf_name)
        return self._create_workflow(workflow_yaml)

    def _create_workflow(self, workflow_yaml):
        yaml_str = pyaml.dump(workflow_yaml)
        workflow_yaml = yaml.safe_load(yaml_str)
        logging.info("Submitting workflow to Argo")
        try:
            response = self._custom_object_api_client.create_namespaced_custom_object(  # noqa: E501
                WorkflowCRD.GROUP,
                WorkflowCRD.VERSION,
                self.namespace,
                WorkflowCRD.PLURAL
                if workflow_yaml["kind"] == WorkflowCRD.KIND
                else CronWorkflowCRD.PLURAL,
                workflow_yaml,
            )
            logging.info(
                'Workflow %s has been submitted in "%s" namespace!'
                % (response.get("metadata", {}).get("name"), self.namespace)
            )
            return response
        except Exception as e:
            logging.error("Failed to submit workflow")
            raise e

    def _create_secret(self, secret_yaml):
        yaml_str = pyaml.dump(secret_yaml)
        secret_yaml = yaml.safe_load(yaml_str)
        return self._core_api_client.create_namespaced_secret(  # noqa: E501
            self.namespace, secret_yaml
        )



"""
Source: https://argoproj.github.io/argo-workflows/fields/#container
"""
from typing import List, Optional

from pydantic import BaseModel

from .envvar import EnvVar
from .resources import Resources


class Container(BaseModel):
    """Container base model"""

    image: str
    command: List[str]
    workingDir: str = "/app"
    args: List[str] = []
    env: Optional[EnvVar] = None
    resources: Optional[Resources] = Resources()



from typing import Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from .columns import Column
from .config import Config


class Node(BaseModel):
    database: str
    schema_: str = Field(alias="schema")
    node_name: str = Field(alias="name")
    resource_type: Literal["model", "test", "macro", "seed"]
    unique_id: str
    alias: str
    config: Config
    description: Optional[str] = None
    columns: Dict[str, Column]
    compiled_code: Optional[str] = None
    depends_on: Dict[str, List[str]]

    @property
    def name(self):
        return str(self.node_name).replace("_", "-").strip("-")

    @property
    def is_model(self):
        return self.resource_type == "model"

    @property
    def is_view(self):
        return self.config.materialized == "view"

    @property
    def is_ephemeral(self):
        return self.config.materialized == "ephemeral"

    @property
    def is_table(self):
        return self.config.materialized == "table"

    @property
    def dependencies(self):
        return self.depends_on.get("nodes", [])



from os.path import join

import click

from dbt_argo.main import main


@click.command(name="dbt-argo", context_settings={"show_default": True})
@click.argument("path", default=".")
@click.option(
    "--target-path",
    default="./target",
    help="""
    Define the dbt target directory location, containing manifest.json.
""",
)
@click.option(
    "--output-file",
    default="dbt-workflow.yaml",
    help="File path to insert output into.",
)
def cli(
    path: str,
    target_path: str,
    output_file: str,
):
    """CLI program to run the dbt-argo converter

    PATH: base project path containing the `.dbt-argo.yml` conf file (default=".")
    """
    config_path = join(path, ".dbt-argo.yml")
    manifest_path = join(path, target_path, "manifest.json")
    main(
        config_path=config_path,
        manifest_path=manifest_path,
        output_path=output_file,
    )


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



from typing import Optional

from pydantic import BaseModel

from .workflow_spec import WorkflowSpec


class CronWorkflowSpec(BaseModel):
    """CronWorkflowSpec reference"""

    suspend: bool = False
    schedule: str
    timezone: str = "Europe/Paris"
    concurrencyPolicy: str = "Replace"
    workflowSpec: Optional[WorkflowSpec] = WorkflowSpec()



from typing import List, Literal, Optional

from pydantic import BaseModel, Field


class Config(BaseModel):
    enabled: bool
    alias: Optional[str] = None
    schema_: Optional[str] = Field(alias="schema")
    database: Optional[str] = None
    tags: List[str]
    materialized: Literal["table", "view", "test", "seed", "ephemeral"]



"""
Source: https://argoproj.github.io/argo-workflows/fields/#resourcerequirements
"""

from pydantic import BaseModel

from .quantity import Limits, Requests


class Resources(BaseModel):
    """ResourceRequirements describes the compute resource requirements."""

    limits: Limits = Limits()
    requests: Requests = Requests()



from typing import List

from .argo.arguments import Arguments
from .argo.container import Container
from .argo.dag_task import DAGTask
from .argo.dag_template import DAGTemplate
from .argo.inputs import Inputs
from .argo.parameter import Parameter
from .argo.template import Template
from .argo.workflow_spec import WorkflowSpec
from .argo.workflow_step import WorkflowStep
from .dbt import DbtManifest, DbtNode

DBT_IMAGE = "ghcr.io/dbt-labs/dbt-bigquery"


class DbtContainer(Container):
    """A single dbt application container to run within a pod."""

    def __init__(self, version: str = "latest"):
        super().__init__(
            image=f"{DBT_IMAGE}:{version}",
            command=["/bin/sh"],
            args=[
                "-c",
                "dbt {{ inputs.parameters.command }} {{ inputs.parameters.args }}",
            ],
        )


class DbtBaseTemplate(Template):
    """dbt Template unit of execution in a workflow"""

    def __init__(self, name="dbt-base", version="latest"):
        super().__init__(
            name=name,
            container=DbtContainer(version=version),
            inputs=Inputs(
                parameters=[Parameter(name="command"), Parameter(name="args", value="")]
            ),
        )


class DbtDAGTask(DAGTask):
    """A single dbt node execution"""

    def __init__(self, node: DbtNode, dependencies: List[str], base_name="dbt-base"):
        super().__init__(
            name=node.name,
            template=base_name,
            arguments=Arguments(
                parameters=[
                    Parameter(name="command", value="run"),
                    Parameter(name="args", value=f"-s {node.node_name}"),
                ]
            ),
            dependencies=dependencies,
        )


class DbtRunTemplate(Template):
    """An Argo template defining the DAG of tasks and dependencies based on
    the dbt manifest given in args"""

    def __init__(self, manifest: DbtManifest, name="run", base_name="dbt-base"):
        tasks = []
        for node in manifest.nodes.values():
            dependencies = manifest.get_dependencies(node)
            dbt_task = DbtDAGTask(node, dependencies, base_name=base_name)
            tasks.append(dbt_task)

        dag_template = DAGTemplate(tasks=tasks)
        super().__init__(name=name, dag=dag_template)


class DbtRunStep(WorkflowStep):
    def __init__(self, name="run"):
        super().__init__(name=name, template=name)


class DbtSeedStep(WorkflowStep):
    def __init__(self, name="seed", base_name="dbt-base"):
        super().__init__(
            name=name,
            template=base_name,
            arguments=Arguments(
                parameters=[
                    Parameter(name="command", value="seed"),
                    # Parameter(name="args", value=""),
                ]
            ),
        )


class DbtPipelineTemplate(Template):
    def __init__(
        self, name="pipeline", base_name="dbt-base", seed_name="seed", run_name="run"
    ):
        dbt_seed_step = DbtSeedStep(name=seed_name, base_name=base_name)
        dbt_run_step = DbtRunStep(name=run_name)

        steps = [
            [dbt_seed_step],
            [dbt_run_step],
        ]

        super().__init__(name=name, steps=steps)


class DbtWorkflowSpec(WorkflowSpec):
    def __init__(
        self,
        manifest: DbtManifest,
        base_name="dbt-base",
        seed_name="seed",
        run_name="run",
        pipeline_name="pipeline",
    ):
        dbt_base_template = DbtBaseTemplate(name=base_name)
        dbt_run_template = DbtRunTemplate(
            manifest, name=run_name, base_name=dbt_base_template.name
        )
        dbt_pipeline_template = DbtPipelineTemplate(
            name=pipeline_name,
            base_name=dbt_base_template.name,
            seed_name=seed_name,
            run_name=run_name,
        )

        templates = [
            dbt_base_template,
            dbt_run_template,
            dbt_pipeline_template,
        ]

        super().__init__(entrypoint=pipeline_name, templates=templates)



from typing import Optional

from pydantic import BaseModel, model_validator


class EnvVar(BaseModel):
    name: str
    value: str = ""
    value_from: Optional[str]

    # https://docs.pydantic.dev/latest/usage/validators/#model-validators
    @model_validator(mode="after")
    def check_value_from(self) -> "EnvVar":
        if self.value != "" and self.value_from is not None:
            raise ValueError("`value_from` cannot be used if `value` is not empty.")
        return self



"""Source: https://argoproj.github.io/argo-workflows/fields/#workflowspec"""
import os
from typing import List, Optional

from pydantic import BaseModel

from .template import Template

_K8S_SA = os.getenv("K8S_SA", "pipeline")


class WorkflowSpec(BaseModel):
    """WorkflowSpec reference"""

    entrypoint: str = "pipeline"
    serviceAccountName: str = _K8S_SA
    templates: Optional[List[Template]] = None



import json
from os import getcwd, getenv, path
from typing import Dict, Optional

from pydantic import BaseModel, field_validator

from .node import Node


class Manifest(BaseModel):
    nodes: Dict[str, Node]

    @field_validator("nodes")
    @classmethod
    def nodes_are_models(cls, nodes: Dict[str, Node]) -> Dict[str, Node]:
        return {node_name: node for node_name, node in nodes.items() if node.is_model}

    @classmethod
    def load(cls, manifest_path: Optional[str] = None):
        """Load a Manifest class from manifest file path."""

        if not manifest_path:
            manifest_path = path.join(
                getenv("DBT_TARGET_DIR", getcwd()), "manifest.json"
            )

        with open(manifest_path, "r", encoding="utf8") as manifest_file:
            manifest_dict = json.load(manifest_file)
        return cls(**manifest_dict)

    def get_dependencies(self, node: Node):
        return [
            self.nodes[dependency].name
            for dependency in node.dependencies
            if dependency in self.nodes and self.nodes[dependency].is_model
        ]



"""
Source: https://argoproj.github.io/argo-workflows/fields/#artifact
"""
from pydantic import BaseModel, Field


class GCSArtifact(BaseModel):
    """GCSArtifact is the location of a GCS artifact"""

    bucket: str = Field(description="Bucket is the name of the bucket")
    key: str = Field(
        description="Key is the path in the bucket where the artifact resides"
    )


class Artifact(BaseModel):
    """Artifact indicates an artifact to place at a specified path"""

    name: str
    path: str = "/app"
    gcs: GCSArtifact



"""
Source: https://argoproj.github.io/argo-workflows/fields/#dagtask
"""

from typing import List

from pydantic import BaseModel

from .dag_task import DAGTask


class DAGTemplate(BaseModel):
    """DAGTemplate is a template subtype for directed acyclic graph templates"""

    tasks: List[DAGTask]



import yaml
from couler.argo_submitter import ArgoSubmitter

from dbt_argo.argo.base import WorkflowFactory
from dbt_argo.dbt import DbtManifest
from dbt_argo.dbt_workflow import DbtWorkflowSpec

from .utils import load_yaml


# T-12 use a "project_dir" folder instead
def main(config_path: str, manifest_path: str, output_path: str):
    """Main function to run to parsing and bumping"""

    config = load_yaml(config_path)
    manifest = DbtManifest.load(manifest_path)

    base_workflow = WorkflowFactory(config)

    dbt_workflow_spec = DbtWorkflowSpec(manifest=manifest)

    dbt_workflow = base_workflow.set_workflow_spec(dbt_workflow_spec)

    with open(output_path, "w", encoding="utf8") as f:
        workflow_yaml = dbt_workflow.model_dump(exclude_none=True)
        yaml.dump(workflow_yaml, f)

    submitter = ArgoSubmitter(namespace="omni-leogrosjean")

    submitter.submit(workflow_yaml)



"""
Source: https://argoproj.github.io/argo-workflows/fields/#template
"""
from typing import List, Optional

from pydantic import BaseModel, model_validator

from .container import Container
from .dag_template import DAGTemplate
from .inputs import Inputs
from .workflow_step import WorkflowStep


class Template(BaseModel):
    """Template is a reusable and composable unit of execution in a workflow"""

    name: str
    container: Optional[Container] = None
    inputs: Optional[Inputs] = None
    dag: Optional[DAGTemplate] = None
    steps: Optional[List[List[WorkflowStep]]] = None

    # https://docs.pydantic.dev/latest/usage/validators/#model-validators
    @model_validator(mode="after")
    def check_template(self) -> "Template":
        if (self.container is None) ^ (self.dag is None) ^ (self.steps is None):
            raise ValueError(
                "`Only one from container, dag or template should be given."
            )
        return self



"""
Source: https://argoproj.github.io/argo-workflows/fields/#inputs
"""
from typing import List, Optional

from pydantic import BaseModel

from .artifact import Artifact
from .parameter import Parameter


class Inputs(BaseModel):
    """Inputs Base model"""

    artifacts: Optional[List[Artifact]] = None
    parameters: Optional[List[Parameter]] = None



import typing as t
from pathlib import Path

from cloudpathlib import GSPath
from google.cloud import storage

DBT_IMAGE = "ghcr.io/dbt-labs/dbt-bigquery"


def clean_name(name: str):
    """
    _int_webinar__webinar_ru > int-webinar--webinar-ru
    """
    return name.replace("_", "-").strip("-")


# Source: https://stackoverflow.com/a/52193880
def copy_dir_to_gcs(
    local_path: t.Union[str, Path],
    gcs_path: str,
    gcp_project_id: t.Optional[str] = None,
):
    """Recursively copy a directory of files to GCS."""

    dir_path = Path(local_path)

    if not dir_path.exists():
        raise NotADirectoryError()

    gcs_pathlib = GSPath(gcs_path)

    bucket_name = gcs_pathlib.bucket

    client = storage.Client(project=gcp_project_id)
    bucket = client.get_bucket(bucket_name)

    for file_path in dir_path.rglob("*"):
        if not file_path.is_file():
            continue

        print(f"Local file: {file_path}")
        relative_file_path = file_path.relative_to(dir_path)
        print(f"Relative file: {relative_file_path}")

        remote_path = gcs_pathlib / relative_file_path.as_posix()

        print(remote_path.blob)
        blob = bucket.blob(remote_path.blob)
        blob.upload_from_filename(file_path)



"""
Source: https://argoproj.github.io/argo-workflows/fields/#arguments
"""

from typing import List

from pydantic import BaseModel

from .parameter import Parameter


class Arguments(BaseModel):
    """Arguments to a template"""

    parameters: List[Parameter] = []



# pylint: disable=line-too-long,too-many-arguments,too-many-locals
import typing as t

from croniter import CroniterBadCronError, croniter
from dbt.contracts.graph.manifest import Manifest
from hera.workflows import (
    DAG,
    Container,
    CronWorkflow,
    Env,
    GCSArtifact,
    Parameter,
    Resources,
    Steps,
    Task,
    Workflow,
)

from .utils import DBT_IMAGE, clean_name


# TODO: creer une fonction "base" qui permet de générer le dbt_base, dag et pipeline
# pour le réutiliser dans dans create_dbt_workflow and create_dbt_cronworkflow
def _base_workflow():
    raise NotImplementedError


def create_dbt_workflow(
    manifest: Manifest,
    gcs_bucket: str,
    env: str,  # Used by dbt to be executed
    gcp_project_id: str,  # Used by dbt to be executed
    command: str = "run",
    name: str = "dbt-workflow",
    service_account_name: t.Optional[str] = None,
    namespace: t.Optional[str] = None,
):
    """
    Create an hera workflow object which runs a dbt pipeline workflow, based on
    dbt Manifest object
    """

    entrypoint = "pipeline"

    with Workflow(
        name=f"{name}",
        namespace=namespace,
        entrypoint=entrypoint,
        service_account_name=service_account_name,
    ) as workflow:
        # create a func dbt_base which returns dbt_base_op
        dbt_base = Container(
            image=DBT_IMAGE,
            working_dir="/app",
            inputs=[
                Parameter(name="command"),  # type: ignore
                Parameter(name="args", default=""),  # type: ignore
                GCSArtifact(name="models", path="/app", bucket=gcs_bucket, key=name),  # type: ignore
            ],
            name="dbt-base",
            command=["/bin/sh"],
            args=["-c", "dbt {{inputs.parameters.command}} {{inputs.parameters.args}}"],
            resources=Resources(
                cpu_limit="0.15",
                cpu_request="0.1",
                memory_limit="300Mi",
                memory_request="200Mi",
            ),
            env=[
                Env(name="ENV", value=env),
                Env(name="GCP_PROJECT_APP", value=gcp_project_id),
                Env(name="DBT_PROFILES_DIR", value="./"),
            ],
        )

        with DAG(name=f"dbt-{command}") as dag:
            tasks = {}

            # #1. create tasks
            for node_name, node in manifest.nodes.items():
                if node.resource_type == "model" and node.config.materialized in (
                    "table",
                    "view",
                    "ephemeral",
                ):
                    # print(node.name)
                    tasks[node_name] = dbt_base(
                        name=clean_name(node.name),
                        arguments={"command": command, "args": f"-s {node_name}"},
                    )

            # 2. Create dependencies
            for node_name, node in manifest.nodes.items():
                if node.resource_type == "model":
                    task = tasks[node_name]
                    for node_parent_name in node.depends_on.nodes:  # type: ignore
                        if not node_parent_name in manifest.nodes:
                            continue
                        node_parent = manifest.nodes[node_parent_name]
                        if node_parent.config.materialized in (
                            "table",
                            "view",
                            "ephemeral",
                        ):
                            task_parent: Task = tasks[node_parent_name]
                            task_parent.next(task)

        with Steps(name=entrypoint):
            dbt_base(name="dbt-seed-step", arguments={"command": "seed"})
            dag(name=f"dbt-{command}-step")

        return workflow


def create_dbt_cronworkflow(
    manifest: Manifest,
    schedule: str,
    gcs_bucket: str,
    env: str,  # Used by dbt to be executed
    gcp_project_id: str,  # Used by dbt to be executed
    command: str = "run",
    name: str = "dbt-workflow",
    service_account_name: t.Optional[str] = None,
    namespace: t.Optional[str] = None,
    timezone: t.Optional[str] = None,
    suspend: t.Optional[bool] = None,
):
    """
    Create an hera workflow object which runs a dbt pipeline workflow, based on
    dbt Manifest object
    """

    if not croniter.is_valid(schedule):
        raise CroniterBadCronError(f"Not valid cron : {schedule}")

    entrypoint = "pipeline"

    with CronWorkflow(
        name=f"{name}",
        schedule=schedule,
        namespace=namespace,
        entrypoint=entrypoint,
        service_account_name=service_account_name,
        timezone=timezone,
        suspend=suspend,
    ) as workflow:
        # create a func dbt_base which returns dbt_base_op
        dbt_base = Container(
            image=DBT_IMAGE,
            working_dir="/app",
            inputs=[
                Parameter(name="command"),  # type: ignore
                Parameter(name="args", default=""),  # type: ignore
                GCSArtifact(name="models", path="/app", bucket=gcs_bucket, key=name),  # type: ignore
            ],
            name="dbt-base",
            command=["/bin/sh"],
            args=["-c", "dbt {{inputs.parameters.command}} {{inputs.parameters.args}}"],
            resources=Resources(
                cpu_limit="0.15",
                cpu_request="0.1",
                memory_limit="300Mi",
                memory_request="200Mi",
            ),
            env=[
                Env(name="ENV", value=env),
                Env(name="GCP_PROJECT_APP", value=gcp_project_id),
                Env(name="DBT_PROFILES_DIR", value="./"),
            ],
        )

        with DAG(name=f"dbt-{command}") as dag:
            tasks = {}

            # #1. create tasks
            for node_name, node in manifest.nodes.items():
                if node.resource_type == "model" and node.config.materialized in (
                    "table",
                    "view",
                    "ephemeral",
                ):
                    # print(node.name)
                    tasks[node_name] = dbt_base(
                        name=clean_name(node.name),
                        arguments={"command": command, "args": f"-s {node_name}"},
                    )

            # 2. Create dependencies
            for node_name, node in manifest.nodes.items():
                if node.resource_type == "model":
                    task = tasks[node_name]
                    for node_parent_name in node.depends_on.nodes:  # type: ignore
                        if not node_parent_name in manifest.nodes:  # remove sources
                            continue
                        node_parent = manifest.nodes[node_parent_name]
                        if node_parent.config.materialized in (
                            "table",
                            "view",
                            "ephemeral",
                        ):
                            task_parent: Task = tasks[node_parent_name]
                            task_parent.next(task)

        with Steps(name=entrypoint):
            dbt_base(name="dbt-seed-step", arguments={"command": "seed"})
            dag(name=f"dbt-{command}-step")

        return workflow



"""
Source: https://argoproj.github.io/argo-workflows/fields/#parameter
"""

from typing import Optional

from pydantic import BaseModel


class Parameter(BaseModel):
    name: str
    # https://docs.pydantic.dev/latest/migration/#required-optional-and-nullable-fields
    value: Optional[str] = None



"""
Source: https://argoproj.github.io/argo-workflows/fields/#workflowstep
"""

from typing import Optional

from pydantic import BaseModel

from .arguments import Arguments


class WorkflowStep(BaseModel):
    """WorkflowStep is a reference to a template to execute in a series of step"""

    name: str
    template: str
    arguments: Optional[Arguments] = None



from typing import Literal

from pydantic import BaseModel

from .metadata import Metadata
from .workflow_spec import WorkflowSpec


class Workflow(BaseModel):
    """Workflow model"""

    kind: Literal["Workflow"]
    metadata: Metadata
    spec: WorkflowSpec

    def set_workflow_spec(self, workflow_spec: WorkflowSpec):
        self.spec = workflow_spec
        return self
